#
# baikal api client tests
#

import pytest

from rstms_bcc import settings
from rstms_bcc.browser import Session
from rstms_bcc.models import (
    Account,
    AddBookRequest,
    AddUserRequest,
    Book,
    DeleteBookRequest,
    DeleteUserRequest,
    User,
)
from rstms_bcc.version import __version__


@pytest.fixture
def test_username():
    return "testo@example.org"


@pytest.fixture
def account(test_env):
    account = Account(username=settings.ADMIN_USERNAME, password=str(settings.ADMIN_PASSWORD))
    return account


@pytest.fixture
def session(account):
    session = Session()
    assert session
    yield session
    session.shutdown()


@pytest.fixture
def test_user(account, session, test_username):
    users = session.users(account)
    for user in users:
        if user.username == test_username:
            return user
    added = session.add_user(
        account, AddUserRequest(username=test_username, displayname="test user", password="testpassword")
    )
    assert isinstance(added, User)
    assert added.username == test_username
    return added


def test_session_status(session):
    status = session.status()
    assert isinstance(status, dict)


def test_session_login(session, account):
    session.login(account)
    assert session.logged_in


def test_session_logout(session, account):
    session.login(account)
    assert session.logged_in
    session.logout()
    assert not session.logged_in


def test_session_users(session, account):
    users = session.users(account)
    assert users
    assert isinstance(users, list)
    for user in users:
        assert isinstance(user, User)


def test_session_add_delete_user(session, account, test_username):
    def delete_testuser():
        users = session.users(account)
        for user in users:
            if user.username == test_username:
                deleted = session.delete_user(account, DeleteUserRequest(username=test_username))
                assert deleted == {"message": f"deleted: {test_username}"}

    delete_testuser()
    added = session.add_user(
        account, AddUserRequest(username=test_username, displayname="test user", password="testpassword")
    )
    assert isinstance(added, User)
    assert added.username == test_username
    delete_testuser()


def test_session_books(session, account, test_user):
    books = session.books(account, test_user.username)
    assert books
    assert isinstance(books, list)
    for book in books:
        assert isinstance(book, Book)


def test_session_add_delete_book(session, account, test_user):

    def delete_testbook():
        books = session.books(account, test_user.username)
        assert books
        assert isinstance(books, list)
        for book in books:
            if book.bookname == "testbook":
                deleted = session.delete_book(account, DeleteBookRequest(username=test_user.username, token=book.token))
                assert deleted == {"message": f"deleted: {book.token}"}

    delete_testbook()
    request = AddBookRequest(username=test_user.username, bookname="testbook", description="test address book")
    added = session.add_book(account, request)
    assert added
    assert isinstance(added, Book)
    assert added.username == test_user.username
    assert added.bookname == "testbook"
    book_names = [b.bookname for b in session.books(account, test_user.username)]
    assert "testbook" in book_names
    delete_testbook()
    book_names = [b.bookname for b in session.books(account, test_user.username)]
    assert "testbook" not in book_names


def test_session_reset(session):
    status = session.status()
    assert isinstance(status, dict)
    assert status["uptime"] is not None
    assert status["version"] == __version__
    ret = session.reset()
    assert ret == dict(message="session reset")
    status = session.status()
    assert isinstance(status, dict)
    assert status["uptime"] is not None
    assert status["reset_time"] is not None
    assert status["version"] == __version__


def test_session_initialized(session, account):
    ret = session.initialize(account)
    assert isinstance(ret, dict)
    assert list(ret.keys()) == ["message"]
    assert ret["message"] in ["already initialized", "configured successfully"]



# test config

import asyncio
import logging
import os
import time
from threading import Thread

import psutil
import pytest

from rstms_bcc import settings
from rstms_bcc.app import app
from rstms_bcc.client import API
from rstms_bcc.models import Book, User

logger = logging.getLogger(__name__)

LISTEN_TIMEOUT = 5

# configure the test environment
TEST_ADDRESS = "127.0.0.1"
TEST_PORT = 8001
TEST_API_KEY = os.environ.get("TEST_API_KEY", "test_api_key")
TEST_API_URL = f"http://{TEST_ADDRESS}:{TEST_PORT}/bcc"
TEST_CALDAV_URL = os.environ["TEST_CALDAV_URL"]
TEST_SERVER_LOG_LEVEL = "INFO"
TEST_USER_PREFIX = "bcc__test__user__"
TEST_BOOK_PREFIX = "bcc__test__book__"

# test control flags
USE_EXISTING_SERVER = "USE_EXISTING_SERVER" in os.environ
PRUNE_TEST_DATA = True  # "NO_PRUNE_TEST_DATA" not in os.environ


@pytest.fixture(scope="session")
def test_env():
    env = os.environ
    try:
        settings.ADDRESS = TEST_ADDRESS
        settings.PORT = TEST_PORT
        settings.API_KEY = TEST_API_KEY
        settings.API_URL = TEST_API_URL
        settings.CALDAV_URL = TEST_CALDAV_URL
        settings.LOG_LEVEL = "INFO"
        settings.DEBUG = False
        # settings.INSECURE_NO_CLIENT_CERT = True
        yield env
    finally:
        for k, v in env.items():
            os.environ[k] = v


"""
@pytest.fixture(scope="session")
def test_api_key():
    return TEST_API_KEY


@pytest.fixture(scope="session")
def test_api_url():
    return TEST_API_URL


@pytest.fixture
def headers():
    return {"X-API-Key": TEST_API_KEY}
"""


class TestServer:
    def __init__(self):
        self.host = settings.ADDRESS
        self.port = settings.PORT
        self.app = app
        self._server = None
        self.running = False

    def run(self):
        import uvicorn

        config = uvicorn.Config(
            app=app,
            host=self.host,
            port=self.port,
            log_level=TEST_SERVER_LOG_LEVEL.lower(),
            loop="asyncio",
            lifespan="on",
        )
        self._server = uvicorn.Server(config)
        self.running = True
        self._server.run()
        self.running = False

    def stop(self):
        if self._server:
            self._server.should_exit = True


def listening_tcp_ports():
    return [conn.laddr.port for conn in psutil.net_connections(kind="tcp") if conn.status == "LISTEN"]


async def not_listening(host, port):
    return await listening(host, port, False)


async def listening(host, port, desired_state=True):
    timeout = time.time() + LISTEN_TIMEOUT
    while True:
        assert time.time() < timeout, f"Timed out waiting for {host}:{port} LISTEN"
        if port in listening_tcp_ports():
            if desired_state is True:
                return
        else:
            if desired_state is False:
                return
        await asyncio.sleep(1)


def prune_test_data(client):
    users = client.users()
    for user in users:
        if user.username.startswith(TEST_USER_PREFIX):
            logger.info(f"PRUNING: {user}")
            client.delete_user(user.username)


@pytest.fixture(scope="session")
async def test_server(test_env):

    if int(TEST_PORT) in listening_tcp_ports():
        assert USE_EXISTING_SERVER, f"server running on port {TEST_PORT}; set USE_EXISTING_SERVER to use"
        yield True
        return
    server = TestServer()
    thread = Thread(target=server.run)
    thread.start()
    try:
        await listening(server.host, server.port)
        yield server
    finally:
        server.stop()
        await not_listening(server.host, server.port)
        thread.join()


@pytest.fixture(scope="session")
def client(test_server):
    client = API()
    yield client
    if PRUNE_TEST_DATA:
        prune_test_data(client)


@pytest.fixture(scope="session")
def testid(client):
    yield str(time.time()).replace(".", "")


@pytest.fixture(scope="function")
def username(testid):
    return f"{TEST_USER_PREFIX}{testid}@domain.ext"


@pytest.fixture(scope="function")
def displayname(testid):
    return f"testuser {testid}"


@pytest.fixture(scope="function")
def password(testid):
    return testid


@pytest.fixture(scope="function")
def bookname(testid):
    return f"{TEST_BOOK_PREFIX}book_{testid}"


@pytest.fixture(scope="function")
def description(testid):
    return f"test book {testid}"


@pytest.fixture(scope="function")
def testuser(client, username, displayname, password):
    """ensure testuser exists"""
    users = client.users()
    for user in users:
        if user.username == username:
            assert user.displayname == displayname
            return user
    user = client.add_user(username, displayname, password)
    assert isinstance(user, User)
    assert user.username == username
    assert user.displayname == displayname
    return user


@pytest.fixture(scope="function")
def testbook(client, testuser, bookname, description):
    """ensure testuser and testbook exist"""
    books = client.books(testuser.username)
    for book in books:
        if book.username == testuser.username and book.bookname == bookname:
            assert book.description == description
            return book
    book = client.add_book(testuser.username, bookname, description)
    assert isinstance(book, Book)
    assert book.username == testuser.username
    assert book.bookname == bookname
    assert book.description == description
    return book



"""Top-level package for bcc."""

import sys
import warnings

if sys.platform.startswith("openbsd"):
    # suppress urllib3 complaint about libreSSL on OpenBSD
    warnings.filterwarnings("ignore", message="urllib3 v2 only supports OpenSSL")

from .app import app
from .cli import bcc
from .version import __author__, __email__, __timestamp__, __version__

__all__ = ["app", "bcc", "__version__", "__timestamp__", "__author__", "__email__"]



import asyncio
import logging
import os
import signal
from contextlib import asynccontextmanager

import arrow
from fastapi import BackgroundTasks, Depends, FastAPI, Header, HTTPException, Request
from fastapi.responses import JSONResponse
from typing_extensions import Annotated

from . import settings
from .browser import BrowserException, Session
from .models import (
    Account,
    AddBookRequest,
    AddBookResponse,
    AddUserRequest,
    AddUserResponse,
    BooksResponse,
    DeleteBookRequest,
    DeleteBookResponse,
    DeleteUserRequest,
    DeleteUserResponse,
    InitializeResponse,
    PasswordResponse,
    ResetResponse,
    ShutdownResponse,
    StatusResponse,
    UptimeResponse,
    UserAccountsResetRequest,
    UserAccountsResponse,
    UsersResponse,
)
from .passwd import AccountException, Accounts
from .version import __version__


async def require_client_cert_header(
    x_client_cert_dn: Annotated[str, Header()],
):
    if x_client_cert_dn != "CN=mabctl":
        raise HTTPException(status_code=401, detail="client certificate CN mismatch")


async def require_api_key_header(
    x_api_key: Annotated[str, Header()],
):
    if x_api_key != app.state.api_key:
        raise HTTPException(status_code=401, detail="invalid API key")


async def require_admin_credential_headers(
    x_admin_username: Annotated[str, Header()],
    x_admin_password: Annotated[str, Header()],
):
    if not x_admin_username:
        raise HTTPException(status_code=401, detail="missing username")
    if not x_admin_password:
        raise HTTPException(status_code=401, detail="missing password")
    app.state.account = Account(username=x_admin_username, password=x_admin_password)


async def null_header_dependency():
    pass


def shutdown_browser_session():
    clear_logout_timer()
    clear_shutdown_timer()
    if app.state.session:
        app.state.session.shutdown()
    app.state.session = None


@asynccontextmanager
async def lifespan(app: FastAPI):
    app.state.logger = logging.getLogger(settings.LOGGER_NAME)
    app.state.logger.setLevel(settings.LOG_LEVEL)
    app.state.logger.info(f"bcc v{__version__} startup")
    app.state.api_key = settings.get(None, "API_KEY", settings.Get.DECODE_SECRET, settings.Get.OPTIONAL_READ_FILE)
    app.state.startup_time = arrow.now()
    app.state.logout_timer_task = None
    app.state.shutdown_timer_task = None
    app.state.account = None
    app.state.session = Session()
    if settings.INSECURE_NO_CLIENT_CERT:
        app.state.logger.warning("INSECURE_NO_CLIENT_CERT: client certificate header validation disabled")
        app.dependency_overrides[require_client_cert_header] = null_header_dependency
    if settings.INSECURE_NO_API_KEY:
        app.state.logger.warning("INSECURE_NO_API_KEY: api key header validation disabled")
        app.dependency_overrides[require_api_key_header] = null_header_dependency
    yield
    app.state.logger.info("shutdown")
    shutdown_browser_session()


app = FastAPI(
    dependencies=[
        Depends(require_client_cert_header),
        Depends(require_api_key_header),
        Depends(require_admin_credential_headers),
    ],
    lifespan=lifespan,
)


def handle_exception(request, exc):
    path = str(request.url)[len(str(request.base_url)) :]
    error = dict(success=False, request=request.method + " /" + path, message=exc.__class__.__name__, detail=exc.args)
    app.state.logger.error(repr(error))
    return JSONResponse(status_code=500, content=error)


@app.exception_handler(BrowserException)
async def browser_exception_handler(request: Request, exc: BrowserException):
    return handle_exception(request, exc)


@app.exception_handler(AccountException)
async def account_exception_handler(request: Request, exc: BrowserException):
    return handle_exception(request, exc)


def clear_shutdown_timer():
    if app.state.shutdown_timer_task:
        app.state.shutdown_timer_task.cancel()
    app.state.shutdown_timer = None


def restart_shutdown_timer(restart=True):
    clear_shutdown_timer()
    if settings.AUTO_SHUTDOWN_TIMER:
        app.state.shutdown_timer_task = asyncio.create_task(auto_shutdown_timer())


async def auto_shutdown_timer():
    await asyncio.sleep(settings.AUTO_SHUTDOWN_TIMER)
    app.state.logger.info("auto browser shutdown")
    app.state.session.shutdown()


def clear_logout_timer():
    """called after each request to logout after timeout"""
    if app.state.logout_timer_task:
        app.state.logout_timer_task.cancel()
    app.state.logout_timer_task = None


def restart_logout_timer():
    clear_logout_timer()
    if settings.AUTO_LOGOUT_TIMER:
        app.state.logout_timer_task = asyncio.create_task(auto_logout_timer())


async def auto_logout_timer():
    await asyncio.sleep(settings.AUTO_LOGOUT_TIMER)
    app.state.logger.info("auto logout")
    app.state.session.logout()


@app.middleware("http")
async def logout_after_request(request: Request, call_next):
    response = await call_next(request)
    restart_logout_timer()
    restart_shutdown_timer()
    return response


@app.get("/bcc/status/")
async def get_status() -> StatusResponse:
    return StatusResponse(request="status", status=app.state.session.status())


@app.post("/bcc/reset/")
async def post_reset() -> ResetResponse:
    return app.state.session.reset()


@app.post("/bcc/initialize/")
async def post_initialize() -> InitializeResponse:
    return app.state.session.initialize(app.state.account)


@app.get("/bcc/users/")
async def get_users() -> UsersResponse:
    return UsersResponse(users=app.state.session.users(app.state.account))


@app.post("/bcc/user/")
async def post_user(request: AddUserRequest) -> AddUserResponse:
    return AddUserResponse(user=app.state.session.add_user(app.state.account, request))


@app.delete("/bcc/user/")
async def delete_user(request: DeleteUserRequest) -> DeleteUserResponse:
    result = app.state.session.delete_user(app.state.account, request)
    Accounts().remove(request.username)
    return result


@app.get("/bcc/books/")
async def get_addressbooks_all() -> BooksResponse:
    users = app.state.session.users(app.state.account)
    books = []
    for user in users:
        books.extend(app.state.session.books(app.state.account, user.username))
    return BooksResponse(books=books)


@app.get("/bcc/books/{username}/")
async def get_addressbooks_user(username: str) -> BooksResponse:
    return BooksResponse(books=app.state.session.books(app.state.account, username))


@app.post("/bcc/book/")
async def post_address_book(request: AddBookRequest) -> AddBookResponse:
    return AddBookResponse(book=app.state.session.add_book(app.state.account, request))


@app.delete("/bcc/book/")
async def delete_book(request: DeleteBookRequest) -> DeleteBookResponse:
    return app.state.session.delete_book(app.state.account, request)


@app.post("/bcc/shutdown/")
async def shutdown(background_tasks: BackgroundTasks) -> ShutdownResponse:
    app.state.logger.warning("received shutdown request")
    shutdown_browser_session()
    background_tasks.add_task(shutdown_app)
    return dict(message="shutdown requested")


@app.get("/bcc/password/{username}/")
async def get_password(username: str) -> PasswordResponse:
    password = Accounts().get(username)
    return PasswordResponse(username=username, password=password)


@app.get("/bcc/accounts/")
async def get_accounts() -> UserAccountsResponse:
    accounts = Accounts()
    response = UserAccountsResponse(message="user accounts", accounts=accounts.read())
    return response


@app.post("/bcc/accounts/")
async def post_accounts(request: UserAccountsResetRequest) -> UserAccountsResponse:
    accounts = Accounts()
    response = UserAccountsResponse(message="user accounts reset", accounts=accounts.write(request.accounts))
    return response


def shutdown_app():
    app.state.logger.warning("shutdown_task: exiting")
    os.kill(os.getpid(), signal.SIGINT)


@app.get("/bcc/uptime/")
async def uptime() -> UptimeResponse:
    return dict(message="started " + app.state.startup_time.humanize(arrow.now()))



"""bcc cli"""

import json
import logging
import sys
from collections.abc import Iterable

import click
import uvicorn

from . import settings
from .app import app
from .client import API
from .exception_handler import ExceptionHandler
from .shell import _shell_completion
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"

ehandler = None


def _ehandler(ctx, option, debug):
    global ehandler
    ehandler = ExceptionHandler(debug)
    return debug


def render(obj):
    if hasattr(obj, "model_dump_json"):
        return json.loads(obj.model_dump_json())
    if isinstance(obj, Iterable) and not (isinstance(obj, (str, bytes, dict))):
        return [render(o) for o in obj]
    else:
        return obj


def output(obj):
    click.echo(json.dumps(render(obj), indent=2))


@click.group("bcc")
@click.version_option(message=header)
@click.option("-d", "--debug", is_eager=True, envvar="DEBUG", is_flag=True, callback=_ehandler, help="debug mode")
@click.option("-u", "--username", help="username (default: admin)")
@click.option("-p", "--password", help="password")
@click.option("--caldav-url", help="caldav server URL")
@click.option("--api-url", help="API server URL")
@click.option("-l", "--log-level", help="server log level (default: WARNING)")
@click.option("-c", "--cert", help="cient certificate file")
@click.option("-k", "--key", help="client certificate key file")
@click.option("-a", "--api-key", help="bcc API key")
@click.option("--passwd-file", help="passwd filename")
@click.option("-v", "--verbose", is_flag=True, help="more output in DEBUG mode")
@click.option("--log-headers", is_flag=True, help="log request, response headers")
@click.option("--log-html", is_flag=True, help="log HTML responses")
@click.option(
    "--shell-completion",
    is_flag=False,
    flag_value="[auto]",
    callback=_shell_completion,
    help="configure shell completion",
)
@click.pass_context
def bcc(  # noqa: C901
    ctx,
    debug,
    username,
    password,
    caldav_url,
    api_url,
    cert,
    key,
    api_key,
    log_level,
    verbose,
    passwd_file,
    shell_completion,
    log_headers,
    log_html,
):
    """bcc - bcc control console"""

    if debug is not None:
        settings.DEBUG = True
    if username is not None:
        settings.USERNAME = username
    if password is not None:
        settings.PASSWORD = password
    if caldav_url is not None:
        settings.CALDAV_URL = caldav_url
    if api_url is not None:
        settings.API_URL = api_url
    if cert is not None:
        settings.CLIENT_CERT = cert
    if key is not None:
        settings.CLIENT_KEY = key
    if api_key is not None:
        settings.API_KEY = api_key
    if log_level is not None:
        settings.LOG_LEVEL = log_level
    if passwd_file is not None:
        settings.PASSWD_FILE = passwd_file
    if verbose is True:
        settings.VERBOSE = verbose
    if log_headers is True:
        settings.DEBUG_LOG_HEADERS = log_headers
    if log_html is True:
        settings.DEBUG_LOG_HTML = log_html

    logging.basicConfig(level=log_level)

    if ctx.invoked_subcommand not in ["config", "server"]:
        ctx.obj = API()


@bcc.command
@click.pass_context
@click.option("--insecure", is_flag=True)
def config(ctx, insecure):
    """show configuration"""
    click.echo(settings.dotenv(reveal_passwords=insecure))
    sys.exit(0)


@bcc.command
@click.pass_obj
def users(ctx):
    """list users"""
    output(ctx.users())


@bcc.command
@click.argument("username")
@click.argument("display-name")
@click.argument("password")
@click.pass_obj
def mkuser(ctx, username, display_name, password):
    """add user account"""
    output(ctx.add_user(username, display_name, password))


@bcc.command
@click.argument("username")
@click.pass_obj
def rmuser(ctx, username):
    """delete user account"""
    output(ctx.delete_user(username))


@bcc.command
@click.argument("username", required=False)
@click.pass_obj
def books(ctx, username):
    """list address books for user"""
    output(ctx.books(username))


@bcc.command
@click.argument("username")
@click.argument("name")
@click.argument("description")
@click.pass_obj
def mkbook(ctx, username, name, description):
    """add address book for user"""
    output(ctx.add_book(username, name, description))


@bcc.command
@click.argument("username")
@click.argument("token")
@click.pass_obj
def rmbook(ctx, username, token):
    """delete address book for user"""
    output(ctx.delete_book(username, token))


@bcc.command
@click.pass_obj
def reset(ctx):
    """restart client driver"""
    output(ctx.reset())


@bcc.command
@click.pass_obj
def initialize(ctx):
    """initialize freshly installed server"""
    output(ctx.initialize())


@bcc.command
@click.pass_obj
def version(ctx):
    """print version number"""
    click.echo(__version__)


@bcc.command
@click.pass_obj
def status(ctx):
    """output status"""
    output(ctx.status())


@bcc.command
@click.pass_obj
def shutdown(ctx):
    """request server shutdown"""
    output(ctx.shutdown())


@bcc.command
@click.pass_obj
def uptime(ctx):
    """server uptime"""
    output(ctx.uptime())


@bcc.command
@click.pass_context
def server(ctx):
    """API server"""

    settings.LOGGER_NAME = "uvicorn"
    uvicorn.run(
        app,
        host=settings.ADDRESS,
        port=settings.PORT,
        log_level=settings.LOG_LEVEL.lower(),
    )


if __name__ == "__main__":
    sys.exit(bcc())  # pragma: no cover



import json
from pathlib import Path

from . import settings


class AccountException(Exception):
    pass


class AccountNotFound(AccountException):
    pass


class Accounts:
    def __init__(self, filename=None):
        if not filename:
            filename = settings.PASSWD_FILE
        try:
            self.path = Path(filename)
            if not self.path.is_file():
                self.write(dict())
                mode = int(settings.PASSWD_MODE, 8)
                self.path.chmod(mode)
        except Exception as e:
            raise AccountException({e.__class__.__name__: str(e)}) from e

    def read(self):
        try:
            return json.loads(self.path.read_text())
        except Exception as e:
            raise AccountException({e.__class__.__name__: str(e)}) from e

    def write(self, values):
        try:
            with self.path.open("w+") as fp:
                fp.write(json.dumps(values, indent=2) + "\n")
        except Exception as e:
            raise AccountException({e.__class__.__name__: str(e)}) from e
        return self.read()

    def get(self, username):
        values = self.read()
        if username in values:
            return values[username]
        raise AccountNotFound(f"{username=}")

    def set(self, username, password):
        values = self.remove(username)
        values[username] = password
        return self.write(values)

    def remove(self, username):
        values = self.read()
        while values.pop(username, None):
            pass
        return self.write(values)



# bcc API client

from typing import Dict, List

import requests
from pydantic import validate_call
from requests.exceptions import JSONDecodeError

from . import settings
from .models import (
    AddBookRequest,
    AddBookResponse,
    AddUserRequest,
    AddUserResponse,
    Book,
    BooksResponse,
    DeleteBookRequest,
    DeleteUserRequest,
    StatusResponse,
    User,
    UsersResponse,
)


class API:
    @validate_call
    def __init__(
        self,
        *,
        url: str | None = None,
        admin_username: str | None = None,
        admin_password: str | None = None,
        client_cert: str | None = None,
        client_key: str | None = None,
        api_key: str | None = None,
    ):
        self.url = settings.get(url, "API_URL").strip("/")

        self.session = requests.Session()
        self.session.cert = (
            settings.get(client_cert, "CLIENT_CERT", settings.Get.VALIDATE_PEM_CERTIFICATE_FILE),
            settings.get(client_key, "CLIENT_KEY", settings.Get.VALIDATE_PEM_PRIVATE_KEY_FILE),
        )
        self.session.headers["X-Admin-Username"] = settings.get(admin_username, "ADMIN_USERNAME")
        self.session.headers["X-Admin-Password"] = settings.get(
            admin_password, "ADMIN_PASSWORD", settings.Get.DECODE_SECRET, settings.Get.OPTIONAL_READ_FILE
        )
        self.session.headers["X-Api-Key"] = settings.get(
            api_key, "API_KEY", settings.Get.DECODE_SECRET, settings.Get.OPTIONAL_READ_FILE
        )

    def _parse_response(self, response):
        if response.ok:
            return response.json()
        try:
            message = response.json()
        except JSONDecodeError:
            message = f"{str(response)} {response.reason}"
        raise RuntimeError(message)

    def _request(self, func, path, **kwargs):
        return self._parse_response(func(f"{self.url}/{path.strip('/')}/", **kwargs))

    def _get(self, path, **kwargs):
        return self._request(self.session.get, path, **kwargs)

    def _post(self, path, **kwargs):
        return self._request(self.session.post, path, **kwargs)

    def _delete(self, path, **kwargs):
        return self._request(self.session.delete, path, **kwargs)

    @validate_call
    def status(self) -> Dict[str, str]:
        response = StatusResponse(**self._get("/status/"))
        return response.status

    @validate_call
    def initialize(self) -> Dict[str, str]:
        return self._post("/initialize/")

    @validate_call
    def reset(self) -> Dict[str, str]:
        return self._post("/reset/")

    @validate_call
    def users(self) -> List[User]:
        response = UsersResponse(**self._get("/users/"))
        return response.users

    @validate_call
    def add_user(self, username: str, displayname: str, password: str) -> User:
        request = AddUserRequest(username=username, displayname=displayname, password=password)
        response = AddUserResponse(**self._post("/user/", data=request.model_dump_json()))
        return response.user

    @validate_call
    def delete_user(self, username: str) -> Dict[str, str]:
        request = DeleteUserRequest(username=username)
        return self._delete("/user/", data=request.model_dump_json())

    @validate_call
    def books(self, username: str | None = None) -> List[Book]:
        if username:
            path = f"/books/{username}/"
        else:
            path = "/books/"
        result = BooksResponse(**self._get(path))
        return result.books

    @validate_call
    def add_book(self, username: str, bookname: str, description: str) -> Book:
        request = AddBookRequest(username=username, bookname=bookname, description=description)
        response = AddBookResponse(**self._post("/book/", data=request.model_dump_json()))
        return response.book

    @validate_call
    def delete_book(self, username: str, token: str) -> Dict[str, str]:
        request = DeleteBookRequest(username=username, token=token)
        return self._delete("/book/", data=request.model_dump_json())

    def shutdown(self):
        return self._post("/shutdown/")

    def uptime(self):
        return self._get("/uptime/")



# models

import string
from typing import Any, Dict, List

from pydantic import BaseModel, Field, ValidationInfo, model_validator

MIN_PASSWORD_LENGTH = 8
VALID_TOKEN_CHARS = string.ascii_lowercase + string.digits + "-"

_re_name = "[a-zA-Z][a-zA-Z0-9\\._%+-]*[a-zA-Z0-9]"
_re_domain = "([a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9]{1,}\\.){1,}[a-zA-Z]{2,}"

regex_name = "^" + _re_name + "$"
regex_email = "^" + _re_name + "@" + _re_domain + "$"
regex_description = "^[a-zA-Z][a-zA-Z0-9@\\._ -]*[a-zA-Z0-9]*$|^$"
regex_password = "^\\S{" + str(MIN_PASSWORD_LENGTH) + ",}$"
regex_token = "^[a-z0-9-]+$|^$"


class Model(BaseModel):

    @model_validator(mode="before")
    @classmethod
    def normalize(cls, value: Any, info: ValidationInfo) -> Any:
        if isinstance(value, dict):
            fields = value.keys()
            for field in fields:
                if value[field] is None:
                    value[field] = ""
                if isinstance(value[field], bytes):
                    value[field] = value[field].decode()
                if isinstance(value[field], str) and field != "password":
                    value[field] = value[field].strip()
                if field in ["username", "bookname", "token"]:
                    value[field] = value[field].lower()
        else:
            raise RuntimeError(f"model_validator: unexpected value type {type(value)} {value=} {info=}")
        return value


class Account(Model):
    username: str = Field(..., pattern=regex_name)
    password: str = Field(..., pattern=regex_password)


class User(Model):
    username: str = Field(..., pattern=regex_email)
    displayname: str | None = Field("", pattern=regex_description)
    uri: str | None = Field("")


class BookBase(Model):
    username: str = Field(..., pattern=regex_email)
    bookname: str = Field(..., pattern=regex_description)
    description: str | None = Field(None, pattern=regex_description)
    contacts: int | None = Field(0)
    uri: str | None = Field("")


class Book(BookBase):
    token: str = Field(None, pattern=regex_token)


class Response(BaseModel):
    success: str | bool | None = Field(True)
    request: str | None = Field("")
    message: str


class AddUserRequest(User):
    password: str = Field(..., pattern=regex_password)


class AddUserResponse(Response):
    request: str | None = Field("add user")
    message: str | None = Field("user added")
    user: User


class PasswordResponse(Response):
    request: str | None = Field("user password")
    message: str | None = Field("password")
    username: str = Field(..., pattern=regex_email)
    password: str = Field(..., pattern=regex_password)


class UserAccountsResetRequest(Model):
    accounts: Dict[str, str] = Field(..., description="key=username, value=password")


class UserAccountsResponse(Response):
    accounts: Dict[str, str] = Field(..., description="key=username, value=password")


class DeleteUserRequest(Model):
    username: str = Field(..., pattern=regex_email)


class DeleteUserResponse(Response):
    request: str | None = Field("delete user")


class AddBookRequest(BookBase):
    pass


class AddBookResponse(Response):
    request: str | None = Field("add address book")
    message: str | None = Field("address book added")
    book: Book


class DeleteBookRequest(Model):
    username: str = Field(..., pattern=regex_email)
    token: str = Field(..., pattern=regex_token)


class DeleteBookResponse(Response):
    request: str | None = Field("delete address book")


class UsersResponse(Response):
    request: str | None = Field("list users")
    message: str | None = Field("user list")
    users: List[User]


class BooksResponse(Response):
    request: str | None = Field("list address books")
    message: str | None = Field("address book list")
    books: List[Book]


class StatusResponse(Response):
    request: str | None = Field("status")
    message: str | None = Field("server status")
    status: Dict[str, Any]


class InitializeResponse(Response):
    request: str | None = Field("initialize")


class ShutdownResponse(Response):
    request: str | None = Field("shutdown")


class UptimeResponse(Response):
    request: str | None = Field("uptime")


class ResetResponse(Response):
    request: str | None = Field("reset")


class ErrorResponse(Response):
    success: bool | None = Field(False)
    message: str | None = Field("RequestFailed")
    detail: str | None = Field("unspecified error")



# baikal api client

# baikal controller client

import logging
import re
import time
import urllib.parse
from typing import Any, Dict, List, Tuple
from uuid import uuid4

import arrow
import httpx
from bs4 import BeautifulSoup
from pydantic import validate_call

from . import settings
from .models import (
    VALID_TOKEN_CHARS,
    Account,
    AddBookRequest,
    AddUserRequest,
    Book,
    DeleteBookRequest,
    DeleteUserRequest,
    User,
)
from .passwd import Accounts
from .version import __version__

LOG_SOUP = False
POST_DELETE_TIMEOUT = 5

TITLE_MAINTENANCE = "Baïkal Maintainance"
ALREADY_INITIALIZED = "Installation was already completed."
INSTALL_SUCCESS_MESSAGE = (
    "Baïkal is now installed, and its database properly configured."
    " For security reasons, this installation wizard is now disabled."
)
JUMBOTRON_INIT_DATABASE = "Baïkal Database setup"
JUMBOTRON_INIT_WIZARD = "Baïkal initialization wizard"
JUMBOTRON_AUTHENTICATION = "Authentication"
JUMBOTRON_LOGIN_SUCCESS = "Dashboard"


class BrowserException(Exception):
    pass


class BrowserInterfaceFailure(BrowserException):
    pass


class InitFailure(BrowserException):
    pass


class AddFailure(BrowserException):
    pass


class DeleteFailure(BrowserException):
    pass


class UnexpectedServerResponse(BrowserException):
    pass


class NoSuchElementException(BrowserException):
    pass


logger = logging.getLogger(__name__)
CLIENT_TIMEOUT = 300
USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:135.0) Gecko/20100101 Firefox/135.0"


class APIError(Exception):
    def __init__(self, response):
        if getattr(response, "is_error", False):
            error = response.json()
            message = error.get("error", repr(response))
            detail = error.get("detail", [])
        else:
            message = repr(response)
            detail = []
        super().__init__(message)
        self.detail = detail


class APIClient:

    def __init__(self, url=None):
        if url is None:
            url = settings.CALDAV_URL
        self.url = url.strip("/")
        self.cert = (settings.CLIENT_CERT, settings.CLIENT_KEY)
        self.client = httpx.Client(cert=(settings.CLIENT_CERT, settings.CLIENT_KEY), follow_redirects=True)
        self.headers = {"X-Api-Key": str(settings.API_KEY), "User-Agent": USER_AGENT}
        self.timeout = CLIENT_TIMEOUT
        self.logger = logger
        self.load_time = time.time()

    def quit(self):
        self.client.close()
        self.client = None

    def _request(self, func, path, **kwargs):
        raise_on_error = kwargs.pop("raise_on_error", True)
        log_errors = kwargs.pop("log_errors", True)
        kwargs.setdefault("timeout", self.timeout)
        boundary = kwargs.pop("multipart_form", None)
        if boundary is not None:
            self.client.headers["Content-Type"] = f"multipart/form-data; boundary={boundary}"
        try:
            self.response = func(self.url + path, **kwargs)
        except httpx.HTTPError as e:
            if raise_on_error:
                raise
            return APIError(e)
        finally:
            self.client.headers.pop("Content-Type", None)
        self.request = self.response.request
        if self.response.is_error:
            ret = APIError(self.response)
            if log_errors:
                self.print_error(ret, logger.error)
            if raise_on_error:
                raise ret
            return ret

        self.logger.debug(f"request: {self.request.method} {self.request.url}")
        if settings.DEBUG_LOG_HEADERS:
            self.logger.debug(f"request_content: {self.response.request.content.decode()}")
            self._log_headers("request", self.response.request.headers)
            self._log_headers("response", self.response.headers)

        self.dom = BeautifulSoup(self.response.text, features="html.parser")

        if settings.DEBUG_LOG_HTML:
            if self.dom:
                lines = self.dom.prettify().split("\n")
                for line in lines:
                    self.logger.debug(line)

        self.loaded_time = time.time()
        return self.dom

    def _log_headers(self, label, headers):
        self.logger.debug(f"{label}_headers:")
        for k, v in dict(headers).items():
            self.logger.debug(f"{k} {v}")

    def get(self, url, **kwargs):
        self.logger.debug(f"GET {url}")
        return self._request(self.client.get, url, **kwargs)

    def post(self, url, **kwargs):
        self.logger.debug(f"POST {url}")
        return self._request(self.client.post, url, **kwargs)

    def patch(self, url, **kwargs):
        self.logger.debug(f"PATCH {url}")
        return self._request(self.client.patch, url, **kwargs)

    def delete(self, url, **kwargs):
        self.logger.debug(f"DELETE {url}")
        return self._request(self.client.delete, url, **kwargs)

    @validate_call
    def post_form(self, path: str, data: Dict[str, Any]):
        boundary = f"----boundary{uuid4().hex}"
        boundary = "----geckoformboundaryccff3ff29bfd70de26178f03ce5668d8"
        body = ""
        for key, value in data.items():
            body += f"--{boundary}\r\n"
            body += f'Content-Disposition: form-data; name="{key}"\r\n\r\n'
            body += f"{value}\r\n"
        body += f"--{boundary}--\r\n"

        ret = self.post(path, content=body.encode("utf-8"), multipart_form=boundary)
        return ret

    @validate_call
    def meta_refresh(self):
        # process meta refresh if present
        if self.dom.find("meta"):
            attrs = self.dom.find("meta").attrs
            if attrs.get("http-equiv", None) == "refresh":
                content = attrs.get("content", ";")
                fields = [c.strip() for c in content.split(";")]
                refresh_time = None
                path = str(self.request.url)
                for i, field in enumerate(fields):
                    if i == 0:
                        refresh_time = self.loaded_time + int(field)
                    elif field.startswith("url="):
                        path = field[4:]
                if refresh_time is None:
                    raise BrowserInterfaceFailure(f"meta refresh delay time missing: {content=}")
                if time.time() < refresh_time:
                    self.logger.warning(f"meta_refresh: waiting; delay={refresh_time - time.time()}")
                    while time.time() < refresh_time:
                        time.sleep(1)
                if self.request.method != "GET":
                    raise BrowserInterfaceFailure(f"unexpected meta refresh method: {self.request.method}")
                self.get(path)

    def response_path(self):
        return str(self.response.url)[len(self.url) :]


class Session:

    def __init__(self):
        self.logger = logging.getLogger(settings.LOGGER_NAME)
        self.logger.setLevel(settings.LOG_LEVEL)
        self.logger.info("Browser session startup")
        self.initialized = None
        self.logged_in = None
        self.driver = None
        self.init_time = arrow.now()
        self.connect_time = None
        self.reset_time = None
        self.startup()

    @validate_call
    def reset(self) -> Dict[str, str]:
        self.logger.debug("request: reset")
        if self.driver:
            self.shutdown()
        self.startup()
        self.reset_time = arrow.now()
        self.logger.debug("request: reset complete")
        return dict(message="session reset")

    @validate_call
    def startup(self) -> Dict[str, str]:
        self.logger.debug("request: startup")
        if self.driver:
            message = "ignoring startup; already started"
            self.logger.warning(message)
        else:
            try:
                self.driver = APIClient()
            except Exception as e:
                raise InitFailure(e)
            self.logged_in = None
            message = "session started"
            self.connect_time = arrow.now()
            if self.initialized is None:
                self.initialized = self.check_initialized()
        self.logger.debug("request: startup complete")
        return dict(message=message)

    def shutdown(self):
        self.logger.debug("request: shutdown")
        if self.driver:
            self.logout()
            self.driver.quit()
            self.driver = None
            self.logged_in = None
            self.connect_time = None
        else:
            self.logger.warning("already shutdown")
        self.logger.debug("shutdown: complete")

    def _normalize_text(self, text, ignore_case):
        if text is not None:
            text = " ".join(text.strip().split())
            if ignore_case:
                text = text.lower()
        return text

    @validate_call
    def _find_element(
        self,
        name: str,
        selector: str,
        *,
        parent: Any | None = None,
        with_text: str | None = None,
        with_classes: List[str] | None = [],
        allow_none: bool | None = False,
        ignore_case: bool | None = False,
        click: bool | None = False,
    ) -> Any:
        if parent is None:
            parent = self.driver.dom
        elements = parent.select(selector)
        if with_text:
            with_text = self._normalize_text(with_text, ignore_case)
            elements = [e for e in elements if self._normalize_text(e.text, ignore_case) == with_text]
        if with_classes:
            with_classes = [self._normalize_text(c, ignore_case) for c in with_classes]
            class_matches = []
            for element in elements:
                element_classes = [self._normalize_text(c, ignore_case) for c in element.attrs.get("class", [])]
                matches = {cls: (cls in element_classes) for cls in with_classes}
                if all(matches.values()):
                    class_matches.append(element)
            elements = class_matches
        if len(elements) == 1:
            return self._click(elements[0], click)
        elif len(elements) > 1:
            raise BrowserInterfaceFailure(
                f"{name} unexpected multiple matches: {selector=} {with_text=} {with_classes=}"
            )
        elif allow_none:
            return None
        raise BrowserInterfaceFailure(f"{name} not found: {selector=} {with_text=} {with_classes=}")

    def _click(self, element, enabled=True):
        if enabled:
            self.driver.get(element.attrs["href"])
        return element

    @validate_call
    def _find_elements(
        self,
        name: str,
        selector: str,
        *,
        parent: Any | None = None,
        with_text: str | None = None,
        allow_none: bool | None = False,
        ignore_case: bool | None = False,
        click: bool | None = False,
    ) -> List[Any]:
        if parent is None:
            parent = self.driver.dom
        elements = parent.select(selector)
        if elements and with_text:
            with_text = self._normalize_text(with_text, ignore_case)
            elements = [element for element in elements if self._normalize_text(element.text, ignore_case) == with_text]
        if elements:
            self._click(elements[0], click)
        elif not allow_none:
            raise BrowserInterfaceFailure(f"{name} no matching elements found: {selector=} {with_text=}")
        return elements

    @validate_call
    def _click_button(self, name: str, selector: str, *, parent: Any | None = None, with_text: str | None = None):
        return self._find_element(name, selector, parent=parent, with_text=with_text, click=True)

    @validate_call
    def _check_popups(self, require_none: bool | None = False, ignore_case: bool | None = False) -> List[str]:
        messages = self.driver.dom.select("body #message")
        ret = [self._normalize_text(m.text, ignore_case) for m in messages]
        if ret and require_none:
            raise UnexpectedServerResponse(repr(ret))
        return ret

    @validate_call
    def login(self, admin: Account):
        self.logger.debug("request: login")

        if self.driver is None:
            self.startup()

        if self.logged_in == admin.username:
            self.logger.debug(f"login: already logged in as {admin.username}")
            return

        if self.initialized is None:
            self.check_initialized()

        if self.initialized is not True:
            raise BrowserInterfaceFailure("server not initialized")

        post_dom = self.driver.post_form("/admin/", dict(auth=1, login=admin.username, password=admin.password))
        self.logger.debug(post_dom.title)
        self._check_popups(require_none=True)
        self.logged_in = admin.username
        self.logger.info(f"authenticated as {admin.username}")

    def normalized_strings(self, element):
        return [s.strip() for s in list(element.strings) if s.strip()]

    def check_initialized(self):
        self.logger.debug(f"check_initialized: driver={self.driver}")
        if self.driver is None:
            ret = None
        else:
            self.driver.get("/admin/install/")
            strings = self.normalized_strings(self.driver.dom)
            first_string = strings[0] if len(strings) > 0 else ""
            self.logger.debug(f"check_initialize: {first_string=}")
            if first_string.startswith(ALREADY_INITIALIZED):
                self.driver.get("/admin/")
                self.initialized = True
                ret = True
            else:
                self.initialized = False
                ret = False
        self.logger.debug(f"check_initialize: {ret=}")
        return ret

    # TODO: read initialization form values from settings
    @validate_call
    def initialize(self, admin: Account) -> Dict[str, str]:

        self.logger.debug("request: initialize")

        if self.driver is None:
            self.startup()

        self.check_initialized()

        if self.initialized is True:
            return dict(message="already initialized")

        if self.initialized is None:
            raise InitFailure("check_initialize failed")

        title_text = self.driver.dom.title.text.strip()
        self.logger.info(f"initialize: {title_text=}")
        if title_text != TITLE_MAINTENANCE:
            raise BrowserInterfaceFailure(f"unexpected page title: {title_text}")

        while True:

            start_button = self._find_element(
                "start button", "body .btn-success", with_text="Start using Baïkal", allow_none=True
            )
            if start_button:
                self.driver.get("/admin/")

            jumbotron = self.normalized_strings(self._find_element("jumbotron", "body .jumbotron"))

            if (JUMBOTRON_AUTHENTICATION in jumbotron) or (JUMBOTRON_LOGIN_SUCCESS in jumbotron):
                if str(self.driver.response.url).endswith("/admin/"):
                    self.initialized = True
                    message = "configured successfully"
                    self.logger.info(f"initialize: {message}")
                    return dict(message=message)
                else:
                    raise InitFailure("unexpected url after start button: {self.driver.response.url}k")

            elif JUMBOTRON_INIT_DATABASE in jumbotron:
                self.logger.info("initialize: database init")
                self._submit_form("database init", {"data[backend]": "sqlite"})
            elif JUMBOTRON_INIT_WIZARD in jumbotron:
                self.logger.info("initialize: installation wizard")
                self._submit_form(
                    "initialization wizard",
                    {
                        "data[timezone]": "UTC",
                        "data[card_enabled]": "1",
                        "data[cal_enabled]": "1",
                        "data[invite_from]": "",
                        "data[dav_auth_type]": "Digest",
                        "data[admin_passwordhash]": admin.password,
                        "data[admin_passwordhash_confirm]": admin.password,
                    },
                )
            else:
                raise InitFailure(f"unrecognzized jumbotron text: {jumbotron}")

        raise BrowserInterfaceFailure("initialization failed")

    def logout(self):
        self.logger.debug("request: logout")
        if self.logged_in:
            self.driver.get("/admin/")
            self._click_navbar_link("logout", "Logout")
            self.logged_in = None
            links = self._navbar_links().keys()
            if list(links) != ["Web Admin"]:
                raise BrowserInterfaceFailure(f"logout: unexpected navbar content after logout: {links}")
            self.logger.debug("logout: logged out")
        else:
            self.logger.debug("logout: not logged in")

    @validate_call
    def _navbar_links(self):
        navbars = self.driver.dom.select("div.navbar")
        if len(navbars) != 1:
            raise BrowserInterfaceFailure("multiple navbars located")
        elements = navbars[0].find_all("a")
        links = {e.text.strip(): e.attrs.get("href", None) for e in elements if e.text.strip()}
        return links

    @validate_call
    def _click_navbar_link(self, name: str, label: str):
        links = self._navbar_links()
        url = links.get(label, None)
        if url:
            self.driver.get(url)
        else:
            raise BrowserInterfaceFailure(f"{name}: navbar link {label} not found in: {list(links.keys())}")

    @validate_call
    def _table_rows(self, name: str, allow_none: bool | None = True):
        table = self._find_element(f"{name} table", "body table", with_classes=[name])
        tbody = self._find_element(f"{name} table body", "tbody", parent=table, allow_none=True)
        if tbody is not None:
            table = tbody
        rows = table.find_all("tr")
        if not rows:
            message = f"no {name} table body rows found"
            if allow_none:
                self.logger.warning(message)
            else:
                raise BrowserInterfaceFailure(message)
        return rows

    @validate_call
    def _parse_user_row(self, row: Any) -> Dict[str, str]:
        col_username = self._find_element("user table row displayname column", "td.col-username", parent=row)
        fields = col_username.find_all(True)
        username = fields[1].text
        mailto = urllib.parse.unquote(fields[3].attrs["href"])
        match = re.match("^mailto:(.*) <(.*)>$", mailto)
        groups = None
        if match:
            groups = match.groups()
        if not groups:
            raise BrowserInterfaceFailure(f"failed decoding user mailto: {mailto}")
        displayname, email = groups
        ret = self._parse_row_info("user", row)
        ret.update(dict(username=username, displayname=displayname, email=email))
        return ret

    @validate_call
    def _parse_book_row(self, row: Any) -> Dict[str, str]:
        ret = {}
        col_displayname = self._find_element("book table row displayname column", "td.col-displayname", parent=row)
        ret["bookname"] = col_displayname.text.strip()
        col_contacts = self._find_element("book table row contacts column", "td.col-contacts", parent=row)
        ret["contacts"] = int(col_contacts.text.strip())
        col_description = self._find_element("book table row description column", "td.col-description", parent=row)
        ret["description"] = col_description.text.strip()
        data = self._parse_row_info("addressbooks", row)
        ret.update(data)
        ret["token"] = ret["uri"].split("/")[-2]
        return ret

    @validate_call
    def _parse_row_info(self, name: str, row: Any) -> Dict[str, str]:
        popover = self._find_element(f"{name} table row actions popover data", "span.popover-hover", parent=row)
        data_content = popover.attrs["data-content"]
        soup = BeautifulSoup(data_content, "html.parser")
        ret = {}
        last_line = None
        for line in soup.strings:
            if last_line == "URI":
                ret["uri"] = line
            elif last_line == "User name":
                ret["username"] = line
            last_line = line
        if "uri" not in ret:
            raise BrowserInterfaceFailure(f"{name} table row info parse failed")
        return ret

    @validate_call
    def _row_action_buttons(self, name: str, row: Any) -> Dict[str, Any]:
        actions = self._find_element(f"{name} table row actions column", "td.col-actions", parent=row)
        action_buttons = self._find_elements(f"{name} table row actions buttons", "a.btn", parent=actions)
        buttons = {e.text.strip(): e for e in action_buttons if e.text}
        return buttons

    @validate_call
    def users(self, admin: Account) -> List[User]:
        self.logger.debug("request: users")
        self.login(admin)
        self._click_navbar_link("users", "Users and resources")
        rows = self._table_rows("users")
        user_list = [User(**self._parse_user_row(row)) for row in rows]
        self.logger.debug(f"users: returning {len(user_list)}")
        return user_list

    @validate_call
    def _submit_form(self, name: str, args: Dict[str, Any]):
        form = self._find_element(f"{name} form", "form")
        fields = self._find_elements(f"{name} fields", "input", parent=form)
        values = {}
        for field in fields:
            if field.attrs["type"] == "hidden":
                values[field.attrs["name"]] = field.attrs["value"]
        for field in [f for f in form.find_all(True) if f.attrs.get("name", "").startswith("data[")]:
            values[field.attrs["name"]] = field.attrs.get("value", None)
        form_fields = list(values.keys())
        for field_name, value in args.items():
            if field_name not in values:
                raise AddFailure(f"field {field_name=} not present in {form_fields=}")
            values[field_name] = value
        for field_name, value in values.items():
            if value is None:
                raise AddFailure(f"no value provided for {field_name=}")
        path = self.driver.response_path()
        self.driver.post_form(path, values)

    @validate_call
    def add_user(self, admin: Account, request: AddUserRequest) -> User:
        self.logger.debug(f"request: add_user {request.username} {request.displayname} ************")
        user = User(**request.model_dump())
        Accounts().set(user.username, request.password)
        self.login(admin)
        self._click_navbar_link("add_user", "Users and resources")
        self._click_button("add user button", "body .btn", with_text="+ Add user")
        self._submit_form(
            "add user",
            {
                "data[username]": user.username,
                "data[displayname]": user.displayname,
                "data[email]": user.username,
                "data[password]": request.password,
                "data[passwordconfirm]": request.password,
            },
        )
        self._check_add_popups("user", f"User {user.username} has been created.")
        _, parsed = self._find_user_row(user.username, allow_none=False)
        added = User(**parsed)
        if added.username == request.username and added.displayname == request.displayname:
            self.logger.debug("add_user: added")
            return added
        raise AddFailure(
            f"added user mismatches request: added={repr(added.model_dump())} request={repr(request.model_dump())}"
        )

    @validate_call
    def _check_add_popups(self, name: str, expected: str):
        popups = self._check_popups()
        self._click_button(f"add {name} close button", "body form .btn", with_text="Close")
        if expected in popups:
            return
        message = f"unexpected add response: {repr(popups)}"
        self.logger.error(message)
        raise AddFailure(message)

    @validate_call
    def delete_user(self, admin: Account, request: DeleteUserRequest) -> Dict[str, str]:
        username = request.username
        self.logger.debug(f"request: delete_user {username}")
        self.login(admin)
        actions = self._find_user_actions(username)
        if not actions:
            return dict(success=False, message=f"user not found: {username}")
        button = actions.get("Delete", None)
        if not button:
            raise BrowserInterfaceFailure("failed to locate Delete button")
        self.driver.get(button.attrs["href"])
        self._find_elements(
            "user delete confirmation button", "div.alert .btn-danger", with_text="Delete " + username, click=True
        )
        self.driver.meta_refresh()
        self.logger.debug("delete_user: deleted")
        Accounts().remove(username)
        return dict(success=True, message=f"deleted: {username}")

    @validate_call
    def _find_user_row(self, username: str, allow_none: bool | None = True) -> Tuple[Any | None, Any | None]:
        self._click_navbar_link("find_user_row", "Users and resources")
        rows = self._table_rows("users", allow_none=allow_none)
        for row in rows:
            user = self._parse_user_row(row)
            if user["username"] == username:
                return row, user
        self.logger.warning(f"user {username} not found")
        if allow_none:
            return None, None
        raise BrowserInterfaceFailure(f"failed to locate user row: {username=}")

    @validate_call
    def _find_user_actions(self, username: str, allow_none: bool | None = True) -> Dict[str, Any]:
        row, _ = self._find_user_row(username, allow_none=allow_none)
        if row:
            return self._row_action_buttons("user", row)
        return {}

    @validate_call
    def _select_user_address_books(self, username: str, allow_none: bool | None = True):
        buttons = self._find_user_actions(username, allow_none=allow_none)
        button = buttons.get("Address Books", None)
        if button:
            url = button.attrs.get("href", None)
            self.driver.get(url)
            return True
        return None

    @validate_call
    def _find_book_row(
        self, username: str, token: str, allow_none: bool | None = True
    ) -> Tuple[Any | None, Dict[str, Any] | None]:
        if not self._select_user_address_books(username, allow_none=allow_none):
            return None, None
        rows = self._table_rows("addressbooks")
        for row in rows:
            parsed = self._parse_book_row(row)
            if parsed["token"] == token:
                return row, parsed
        return None, None

    @validate_call
    def books(self, admin: Account, username: str) -> List[Book]:
        self.logger.debug(f"request: books {username}")
        self.login(admin)
        if not self._select_user_address_books(username):
            self.logger.debug("books: returning 0")
            return []
        rows = self._table_rows("addressbooks")
        ret = [Book(**self._parse_book_row(row)) for row in rows]
        self.logger.debug(f"books: returning {len(ret)}")
        return ret

    @validate_call
    def add_book(self, admin: Account, request: AddBookRequest) -> Book:
        self.logger.debug(f"request: add_book {request.username} {request.bookname} {request.description}")
        self.login(admin)
        user_row, _ = self._find_user_row(request.username)
        if user_row is None:
            raise AddFailure(f"user not found: username={request.username}")
        if request.bookname == "default address book":
            token = "default"
        else:
            token = request.username + "-" + request.bookname
        token = "".join([c if c in VALID_TOKEN_CHARS else "-" for c in token])
        book = Book(token=token, **request.model_dump())

        row, _ = self._find_book_row(book.username, book.token)
        if row is not None:
            raise AddFailure(f"address book exists: username={book.username} token={book.token}")

        self._click_button("add address book button", "body .btn", with_text="+ Add address book")
        self._submit_form(
            "add book",
            {
                "data[uri]": book.token,
                "data[displayname]": book.bookname,
                "data[description]": book.description,
            },
        )
        self._check_add_popups("addressbook", f"Address Book {book.bookname} has been created.")

        _, parsed = self._find_book_row(request.username, token, allow_none=False)
        added = Book(**parsed)
        if (
            added.username == request.username
            and added.bookname == request.bookname
            and added.description == request.description
            and added.token == token
        ):
            self.logger.debug("add_book: added")
            return added
        raise AddFailure(
            f"added book mismatches request: added={repr(added.model_dump())} request={repr(request.model_dump())}"
        )

    @validate_call
    def delete_book(self, admin: Account, request: DeleteBookRequest) -> Dict[str, str]:
        self.logger.debug(f"request: delete_book {request.username} {request.token}")
        self.login(admin)
        user_row, _ = self._find_user_row(request.username)
        if user_row is None:
            raise DeleteFailure(f"user not found: username={request.username}")
        row, book = self._find_book_row(request.username, request.token)
        if not row:
            raise DeleteFailure(f"book not found: username={request.username} token={request.token}")
        actions = self._row_action_buttons("addressbook", row)
        button = actions.get("Delete", None)
        if not button:
            raise BrowserInterfaceFailure("failed to locate address book Delete button")
        self.driver.get(button.attrs["href"])
        self._find_elements(
            "book delete confirmation button",
            "div.alert .btn-danger",
            with_text="Delete " + book["bookname"],
            click=True,
        )
        self.driver.meta_refresh()
        self.logger.debug("delete_book: deleted")
        return dict(message=f"deleted: {request.token}")

    @validate_call
    def status(self) -> Dict[str, str]:
        self.logger.debug("request: status")
        state = dict(
            name="bcc",
            version=__version__,
            url=settings.CALDAV_URL,
            uptime=self.init_time.humanize(),
            reset_time=self.reset_time.humanize() if self.reset_time else None,
            certificates=repr(self.driver.cert) if self.driver else None,
            connect_time=self.connect_time.humanize() if self.driver else None,
            connected=self.driver is not None,
            initialized=self.initialized,
            login=self.logged_in,
        )
        self.logger.debug("status: returning state")
        return state



# cli tests

import json
import os
import shlex

import pytest
from click.testing import CliRunner

import rstms_bcc as bcc_module
from rstms_bcc import __version__, bcc
from rstms_bcc.models import Book, User


@pytest.fixture(scope="module")
def run(client):
    runner = CliRunner()

    env = os.environ.copy()
    env["TESTING"] = "1"

    def _run(cmd, **kwargs):
        assert_exit = kwargs.pop("assert_exit", 0)
        assert_exception = kwargs.pop("assert_exception", None)
        parse_json = kwargs.pop("parse_json", True)
        env.update(kwargs.pop("env", {}))
        kwargs["env"] = env
        result = runner.invoke(bcc, cmd, **kwargs)
        if assert_exception is not None:
            assert isinstance(result.exception, assert_exception)
        elif result.exception is not None:
            raise result.exception from result.exception
        elif assert_exit is not None:
            assert result.exit_code == assert_exit, (
                f"Unexpected {result.exit_code=} (expected {assert_exit})\n"
                f"cmd: '{shlex.join(cmd)}'\n"
                f"output: {str(result.output)}"
            )
        if parse_json:
            return json.loads(result.output)
        return result

    yield _run


def test_cli_version():
    assert bcc_module.__name__ == "rstms_bcc"
    assert __version__
    assert isinstance(__version__, str)


def test_cli_status(run):
    result = run(["status"])
    assert isinstance(result, dict)
    print(result)


def test_cli_mkuser(run, testuser, password, client):
    delete_result = client.delete_user(testuser.username)
    assert delete_result == dict(request="delete user", success=True, message=f"deleted: {testuser.username}")
    result = run(["mkuser", testuser.username, testuser.displayname, password])
    assert isinstance(result, dict)
    user = User(**result)
    assert isinstance(user, User)
    assert user.username == testuser.username
    assert user.displayname == testuser.displayname
    users = client.users()
    assert testuser in users


def test_cli_users(run, testuser):
    result = run(["users"])
    assert len(result)
    assert isinstance(result, list)
    for r in result:
        assert isinstance(r, dict)
    users = [User(**r) for r in result]
    assert testuser in users


def test_cli_rmuser(run, testuser, client):
    result = run(["rmuser", testuser.username])
    assert result == dict(request="delete user", success=True, message=f"deleted: {testuser.username}")
    users = client.users()
    assert testuser not in users


def test_cli_mkbook(run, testuser, testbook, client):
    assert testuser.username == testbook.username
    delete_result = client.delete_book(testbook.username, testbook.token)
    assert delete_result == dict(request="delete address book", success=True, message=f"deleted: {testbook.token}")
    result = run(["mkbook", testuser.username, testbook.bookname, testbook.description])
    assert isinstance(result, dict)
    book = Book(**result)
    assert isinstance(book, Book)
    assert book.username == testbook.username
    assert book.bookname == testbook.bookname
    assert book.description == testbook.description
    assert book.token


def test_cli_books(run, testbook):
    results = run(["books"])
    assert isinstance(results, list)
    for result in results:
        assert isinstance(result, dict)
    books = [Book(**result) for result in results]
    assert testbook in books


def test_cli_rmbook(run, testuser, testbook, client):
    assert testuser.username == testbook.username
    result = run(["rmbook", testuser.username, testbook.token])
    assert result == dict(request="delete address book", success=True, message=f"deleted: {testbook.token}")
    books = client.books(testuser.username)
    assert testbook not in books



import json
import stat
from pathlib import Path

import pytest

from rstms_bcc import settings
from rstms_bcc.passwd import Accounts


def _chmod(file, mode):
    Path(file).chmod(int(mode, 8))


def _mode(file):
    mode = stat.S_IMODE(Path(file).stat().st_mode)
    return oct(mode)[2:].zfill(4)


def test_passwd_create(shared_datadir):
    file = shared_datadir / "passwd.bcc"
    assert file.is_file() is False
    settings.PASSWD_FILE = str(file)
    passwd = Accounts()
    assert passwd
    with pytest.raises(KeyError):
        passwd.get("howdy")
    assert file.is_file() is True
    assert _mode(file) == "0660"
    assert settings.PASSWD_MODE == "0660"


def test_passwd_exists(shared_datadir):
    file = shared_datadir / "passwd.bcc"
    file.write_text(json.dumps(dict(howdy="testvalue")))
    file.chmod(int("0600", 8))
    settings.PASSWD_FILE = str(file)
    settings.PASSWD_MODE = "0600"
    passwd = Accounts()
    assert passwd
    assert passwd.path == file
    assert passwd.read() == dict(howdy="testvalue")
    assert stat.S_IMODE(file.stat().st_mode) == int("0600", 8)
    assert passwd.get("howdy") == "testvalue"


def test_passwd_change(shared_datadir):
    file = shared_datadir / "passwd.bcc"
    file.write_text(json.dumps(dict(howdy="testvalue")))
    _chmod(file, "0664")
    assert _mode(file) == "0664"
    settings.PASSWD_FILE = str(file)
    settings.PASSWD_MODE = "0600"
    passwd = Accounts()
    assert passwd
    assert passwd.path == file
    assert _mode(file) == "0664"
    assert passwd.read() == dict(howdy="testvalue")
    assert _mode(file) == "0664"
    assert passwd.get("howdy") == "testvalue"
    passwd.set("howdy", "changedvalue")
    assert passwd.get("howdy") == "changedvalue"
    assert passwd.path == file
    assert _mode(file) == "0664"



# settings

import os
import re
import socket
import subprocess
from enum import Enum
from pathlib import Path
from typing import Any

from pydantic import validate_call
from starlette.config import Config
from starlette.datastructures import Secret


class Get(Enum):
    OPTIONAL_READ_FILE = 1
    VALIDATE_PEM_CERTIFICATE_FILE = 2
    VALIDATE_PEM_PRIVATE_KEY_FILE = 3
    VALIDATE_PEM_PUBLIC_KEY_FILE = 4
    DECODE_SECRET = 5


default_webdriver_bin = subprocess.check_output("which geckodriver 2>/dev/null || true", shell=True, text=True).strip()
default_firefox_bin = subprocess.check_output("which firefox 2>/dev/null || true", shell=True, text=True).strip()

###############################################
#
#   Config file precedence: (dotenv format)
#
#    1 env var BCC_CONFIG_FILE
#    2 ./.bcc
#    3 $HOME/.config/bcc/config
#    4 /etc/bcc/config
#
###############################################

for CONFIG_FILE in [
    Path(os.environ.get("BCC_CONFIG_FILE", "")),
    Path(".bcc"),
    Path.home() / ".config" / "bcc" / "config",
    Path("/etc/bcc/config"),
]:
    if CONFIG_FILE.is_file():
        break
config = Config(CONFIG_FILE)

ADDRESS = config("ADDRESS", cast=str, default="127.0.0.1")
PORT = config("PORT", cast=int, default=8000)

ADMIN_USERNAME = config("ADMIN_USERNAME", cast=str, default="admin")
ADMIN_PASSWORD = config("ADMIN_PASSWORD", cast=Secret)
API_KEY = config("API_KEY", cast=Secret)
CALDAV_URL = config(
    "CALDAV_URL", cast=str, default="https://caldav." + ".".join(socket.getfqdn().split(".")[1:]) + "/baikal"
)
API_URL = config("API_URL", cast=str, default="https://caldav." + ".".join(socket.getfqdn().split(".")[1:]) + "/bcc")

CLIENT_CERT = config("CLIENT_CERT", cast=str, default=str(Path.home() / "certs" / "client.pem"))
CLIENT_KEY = config("CLIENT_KEY", cast=str, default=str(Path.home() / "certs" / "client.key"))

INSECURE_NO_CLIENT_CERT = config("INSECURE_NO_CLIENT_CERT", cast=bool, default=False)
INSECURE_NO_API_KEY = config("INSECURE_NO_API_KEY", cast=bool, default=False)

AUTO_LOGOUT_TIMER = config("AUTO_LOGOUT_TIMER", cast=int, default=5)
AUTO_SHUTDOWN_TIMER = config("AUTO_SHUTDOWN_TIMER", cast=int, default=30)

DEBUG = config("DEBUG", cast=bool, default=False)
LOG_LEVEL = config("LOG_LEVEL", cast=str, default="WARNING")
LOGGER_NAME = config("LOGGER_NAME", cast=str, default="bcc")
VERBOSE = config("VERBOSE", cast=bool, default=False)
DEBUG_LOG_HTML = config("DEBUG_LOG_HTML", cast=bool, default=False)
DEBUG_LOG_HEADERS = config("DEBUG_LOG_HEADERS", cast=bool, default=False)

PASSWD_FILE = config("PASSWD_FILE", cast=str, default=str(Path(CONFIG_FILE).parent / "passwd"))
PASSWD_MODE = config("PASSWD_MODE", cast=str, default="0660")


def dotenv(reveal_passwords=False):
    ret = ""
    sep = ""
    for key in sorted([k for k in globals().keys() if re.match("^[A-Z][A-Z_]*$", k)]):
        value = globals()[key]
        if isinstance(value, Secret):
            if reveal_passwords:
                value = str(value)
            else:
                value = "****************"
        if (not isinstance(value, str)) and str(value).isnumeric():
            value = str(value)
        if value is True:
            value = "1"
        elif value is False:
            value = "0"
        elif isinstance(value, str):
            if " " in value or "=" in value or '"' in value or "'" in value:
                value = "'" + value + "'"
        ret += f"{sep}{key}={value}"
        sep = "\n"
    return ret


def read_secret(value):
    if isinstance(value, Secret):
        value = str(value)
    if not value.startswith("@"):
        return value
    value = value[1:]
    if value.startswith("~"):
        path = Path.home()
        value = value[1:]
    else:
        path = Path(".")
    file = path / value
    value = file.read_text()
    value = value.strip()
    return value


def validate_pem_file(filename: str, pem_type: str):
    with Path(filename).open("r") as ifp:
        content = ifp.read()
        if "-----BEGIN" not in content or "-----END" not in content:
            raise ValueError(f"{filename} is not PEM format")
        if "cert" in pem_type:
            if not re.match(".*-----BEGIN CERTIFICATE-----.*", content, re.MULTILINE):
                raise ValueError(f"{filename} is not a certificate")
        elif "key" in pem_type:
            if "pub" in pem_type:
                if not re.match(".*-----BEGIN .*PUBLIC KEY-----.*", content, re.MULTILINE):
                    raise ValueError(f"{filename} is not a public key")
            elif "priv" in pem_type:
                if not re.match(".*-----BEGIN .*PRIVATE KEY-----.*", content, re.MULTILINE):
                    raise ValueError(f"{filename} is not a private key")
            else:
                if not re.match(".*-----BEGIN .* KEY-----.*", content, re.MULTILINE):
                    raise ValueError(f"{filename} is not a key")


@validate_call
def get(value: Any | None, name: str, *flags: Get) -> Any:
    """if value is None, set it from the named setting, performing post-processing if flags are set"""
    if value is None:
        value = globals()[name]
    for flag in flags:
        if flag == Get.DECODE_SECRET:
            if isinstance(value, Secret):
                value = str(value)
        elif flag == Get.OPTIONAL_READ_FILE:
            value = read_secret(value)
        elif flag == Get.VALIDATE_PEM_CERTIFICATE_FILE:
            validate_pem_file(value, "certificate")
        elif flag == Get.VALIDATE_PEM_PRIVATE_KEY_FILE:
            validate_pem_file(value, "privkey")
        elif flag == Get.VALIDATE_PEM_PUBLIC_KEY_FILE:
            validate_pem_file(value, "pubkey")
        else:
            raise ValueError(f"unrecognized settings flag: {flag}")
    return value



import pytest
from pydantic import ValidationError

from rstms_bcc.models import (
    Account,
    AddBookRequest,
    AddUserRequest,
    Book,
    DeleteBookRequest,
    DeleteUserRequest,
    User,
)


@pytest.fixture
def valid_emails():
    return [
        "name@domain.ext",
        "name@one.domain.ext",
        "name@one.two.domain.ext",
        "name-dot@domain.ext",
        "name-dash@domain.ext",
        "name%percent@domain.ext",
        "name+plus@domain.ext",
    ]


@pytest.fixture
def invalid_emails():
    return [
        "plainaddress",  # Missing '@' and domain
        "@missingusername.com",  # Missing username
        "missingatsign.com",  # Missing '@'
        "username@.com",  # Domain missing before the dot
        "username@domain..com",  # Consecutive dots in the domain
        "username@domain.c",  # TLD is too short (1 character)
        "user@domain-.com",  # Hyphen at the end of a domain
        "user@.domain.com",  # Dot immediately after the '@'
        "user@domain.com.",  # Trailing dot in the email
        "user@domain..com",  # Consecutive dots in the domain part
    ]


@pytest.fixture
def invalid_passwords():
    return ["short", " startspace", "endspace ", "embedded space", "", 1, None]


def test_models_account(invalid_passwords):
    account = Account(username="username", password="password")
    assert isinstance(account, Account)
    with pytest.raises(ValidationError):
        _ = Account()
    with pytest.raises(ValidationError):
        _ = Account(username="ok")
    with pytest.raises(ValidationError):
        _ = Account(password="password")
    for password in invalid_passwords:
        with pytest.raises(ValidationError):
            _ = Account(username="valid", password=password)


def test_models_user(valid_emails, invalid_emails):
    for email in valid_emails:
        user = User(username=email, displayname="user name")
        assert isinstance(user, User)
        no_display = User(username=email)
        assert isinstance(no_display, User)

    for bad_email in invalid_emails:
        with pytest.raises(ValidationError):
            _ = User(username=bad_email)


def test_models_book():
    book = Book(username="name@domain.com", bookname="address_book_name", description="address book description")
    assert isinstance(book, Book)


def test_models_add_user():
    request = AddUserRequest(username="name@domain.com", displayname="user name", password="password")
    assert isinstance(request, AddUserRequest)


def test_models_delete_user():
    request = DeleteUserRequest(username="name@domain.com")
    assert isinstance(request, DeleteUserRequest)


def test_models_add_book():
    request = AddBookRequest(
        username="name@domain.com", bookname="address_book_name", description="address book description"
    )
    assert isinstance(request, AddBookRequest)


def test_models_delete_book():
    request = DeleteBookRequest(username="name@domain.com", token="this-is-a-token-1234")
    assert isinstance(request, DeleteBookRequest)
